/***********************************************************************
 * SUMUS.C
 *  Servidor de partidas de mus.
 *  Autor: Antonio Garcia
 *         angarcia@openbank.es
 *  Copyright (C) 2000  Antonio Garcia
 *
 *  Este programa es Software Libre. Puedes usarlo, redistribuirlo o
 *  modificarlo siempre que se respeten los terminos de la licencia
 *  GNU GPL version 2 o (a su eleccion) otra posterior, publicada por
 *  la Free Software Foundation.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 02/01/00 - Consigo completar el diagrama de estados.
 * 07/01/00 - Siete dias despues del "supuesto" efecto 2K empiezo a
 *            programar.
 * 14/01/00 - El protocolo MPP esta casi a punto
 * 02/04/00 - Por fin acabo servidor y cliente para pruebas. Falta Debug.
 * 27/04/00 - Incluyo mensajes de Pares SI/NO y Juego SI/NO. Cambio rutinas
 *	      de cambio de baza y tanteo que no funcionaban ni "patras".
 * 30/04/00 - Doy por buena y acabada la primera version del Servidor. Version 0.1
 * 30/05/00 - Cambios para portabilidad a Win32 (CYGWIN) Version 0.1.5
 * 01/07/00 - Completo un mini-servidor web para servir paginas (y applets java) Version 0.2.1
 * 08/07/00 - Se contempla la 31 Real que se gana con 3 sietes y cualquier sota.
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>

/* SON POCAS DECLARACIONES ASI QUE VAN AQUI EN VEZ DE UN .h APARTE */

/* Definicion de las bazas */

#define REGISTRO 0
#define PREGUNTOMUS 1
#define DESCARTE 2
#define GRANDE   3
#define CHICA    4
#define PARES    5
#define JUEGO    6
#define PUNTO    7

/* Definicion de los mensajes */

struct sMensaje
{
 char id ;
 union TipoMensaje
 {
  struct C_CREAPART
  {
   char NombrePartida[16] ;
   char Vacas ;
   char JuegosPorVaca ;
   char PuntosPorJuego[2] ;
   char Privada ;
   char Jugador[16] ;
   char Companero[16] ;
   char Jugador1[16] ;
   char Jugador3[16] ;
  } c_CreaPart ;
  struct RESPART
  {
   char SeAcepta ;
   char Texto[32] ;
  } ResPart ;
  struct C_APUNPART
  {
   char NombrePartida[16] ;
   char NombreJugador[16] ;
  } c_ApunPart ;
  struct PARTCOMP
  {
   char Vacas ;
   char JuegosPorVaca ;
   char PuntosPorJuego[2] ;
   char NumeroJugador ;
   char Jugador0[16] ;
   char Jugador1[16] ;
   char Jugador2[16] ;
   char Jugador3[16] ;
  } PartComp ;
  struct C_CHAT
  {
   char Texto[64] ;
  } c_Chat ;
  struct S_CHAT
  {
   char Jugador[16] ;
   char Texto[64] ;
  } s_Chat ;
  struct C_PARTCANC
  {
   char nada ;
  } c_PartCanc ;
  struct S_PARTCANC
  {
   char Texto[64] ;
  } s_PartCanc ;
 } a ;
 char SaltodeLinea ;
} *Mensaje ;


struct sPartida
{
 char NombrePartida[17] ; /* Nombres son 16 caracteres + cerito */
 int EsPublica ;          /* Se puede apuntar cualquiera o no */
 char Jugadores[4][17] ;  /* Jugadores 0-2 son pareja A. */
 int Canutos[4] ;         /* socket para comunicacion TCP con jugadores*/
 int CuantasVacas ;       /* Creo que las variablen siguientes se explican solas */
 int JuegosPorVaca ;
 int PuntosPorJuego ;
 int BazaActual ;
 int PuntosActual[2] ;
 int JuegosActual[2] ;
 int VacasActual[2] ;
 int Envidado ;
 int HayOrdago ;
 int PorSumar ;
 int Mano ;
 int QuienHabla ;
 int QuienEnvido ;
 int EnvidadoFinal[8] ; /* PARA EL CONTEO DE FIN DE JUEGO */
 int QuienGanoPares ;   /* Idem. */
 int QuienGanoJuego ;   /* Idem. */
 int Cartas [4][4] ;    /* Las cartas de cada jugador 1=AsOros,40=ReyBastos */
 int Mazo[40] ;		/* EL MAZO DE CARTAS */
 int PorDondeVaElMazo ; /* PUNTERO A LA SIGUIENTE CARTA */
 time_t UltimaAccion ;  /* PARA FINALIZAR UNA PARTIDA POR INACTIVIDAD */
 int Anterior ;		/* PARA HACER UNA LISTA DOBLEMENTE ENLAZADA CON EL ARRAY */
 int Siguiente ;
} *dummyya ;

/* DEFINICIONES PARA PORTABILIDAD */

#ifndef MSG_NOSIGNAL
#define MSG_NOSIGNAL 0
#endif
#ifndef FIONBIO
#define FIONBIO            0x5421
#endif

/* Algunas variables globales */

long bloquea = 1 ;

int MaxPartidas ;
int PuertoWeb ;
int PuertoMus ;
int PosPrimeraPartida = -1 ;
struct sPartida *Partidas ;
int NumPartidas ;
int SocketWeb, SocketMus ;
int SirveWeb = 0 ;
#define MAXWEBPEND 200 /* TAMAO LISTAS DE SOCKETS EN ESPERA DE SER SERVIDOS */
#define MAXMUSPEND 200
#define TIMEOUT 600  /* SI EN 10 Minutos NO SE HA HECHO NADA SE CIERRA LA PARTIDA */
int SocketWebPendientes[MAXWEBPEND] ; /* Conexiones al puerto WEB de las que espero la peticion http */
int CuantosWebPendientes=0 ;
int SocketMusPendientes[MAXMUSPEND] ; /* Conexiones al puerto MUS sin partida asignada todavia */
int CuantosMusPendientes=0 ;

char *CabeceraHTTP = "HTTP/1.0 200 OK\nContent-Type: text/html\n\n<html><head><title>Servidor de partidas de mus</title></head>
<body><H1><Center> Servidor de Partidas de Mus</H1><HR></Center>" ;
char *FinalHTTP = "<br><hr></center>Programado por: Antonio Garcia (<A HREF=\"mailto:angarcia@openbank.es\">angarcia@openbank.es</A>)</body></html>" ;
char Buffer[65536] ;   /* UN PAR DE BUFFERS QUE SIRVEN PARA VARIAS COSAS */
char Buffer2[20480] ;  /* Y SIEMPRE SON MUY SOCORRIDOS A LA HORA DE PROGRAMAR */
char BufferEnv[1024] ;

/************** FUNCIONES DE APOYO *********************/

void Envia( int Canuto, char *Texto )
{
 memset( BufferEnv, ' ', sizeof( struct sMensaje ) ) ;
 memcpy( BufferEnv, Texto, strlen( Texto ) ) ;
 BufferEnv[ sizeof( struct sMensaje ) -1 ] = '\n' ;
 send( Canuto, BufferEnv, sizeof( struct sMensaje ), MSG_NOSIGNAL ) ; 
}

void EliminaPartida( int Pos )
{
 int j ;  /* SI ERA LA PRIMERA PARTIDA DE LA LISTA, QUITARLA DE AHI */
 printf( "Partida eliminada %d [%s]\n",Pos,Partidas[Pos].NombrePartida ) ;
 if( PosPrimeraPartida == Pos )
   PosPrimeraPartida = Partidas[Pos].Siguiente ;
  /* SACARLA DE LA LISTA ENLAZADA DE PARTIDAS */
 if( Partidas[Pos].Anterior > -1 )
   Partidas[ Partidas[Pos].Anterior ].Siguiente = Partidas[Pos].Siguiente ;
 if( Partidas[Pos].Siguiente > -1 )
   Partidas[ Partidas[Pos].Siguiente ].Anterior = Partidas[Pos].Anterior ;
  /* INICIALIZAR EL REGISTRO PARA UNA PROXIMA PARTIDA */
 for( j=0 ; j<4 ; j++ )
 {
  close( Partidas[Pos].Canutos[j] ) ; /* CIERRO LOS SOCKETS */
  Partidas[Pos].Canutos[j] = -1 ;
  Partidas[Pos].Jugadores[j][0] = 0 ;
 }
 Partidas[Pos].NombrePartida[0] = 0 ;
 NumPartidas-- ;
}

/* PreguntaAccion elabora y envia el mensaje "A" (pregunta si envida, pasa,etc) al jugador que habla */
void PreguntaAccion( int Pos )
{
 int j ;

  /* Por alguna razon se llama a esta funcion tras acabar la ronda */
 if( Partidas[Pos].BazaActual > PUNTO || Partidas[Pos].BazaActual < GRANDE ) return ; 

 sprintf( Buffer,"A%1d%1d%1d%1d%1d%1d%1d%2d%2d",Partidas[Pos].QuienHabla,Partidas[Pos].BazaActual-2,
	Partidas[Pos].Envidado > 0 ? 0:1, Partidas[Pos].Envidado > 0 && Partidas[Pos].HayOrdago == 0 ? 1:0,
	Partidas[Pos].Envidado > 0 ? 1:0, Partidas[Pos].HayOrdago == 0 ? 1 : 0, Partidas[Pos].Envidado,
	Partidas[Pos].PorSumar ) ;
 for( j=0 ; j<4 ; j++ )
   Envia( Partidas[Pos].Canutos[j], Buffer ) ;
}

/* MensajeTrasAccion envia el mensaje E a todos los jugadores cuando alguien hablo
   indicando quien hablo, que dijo (Envido, paso,etc) y como queda la cosa */
void MensajeTrasAccion( int Quien, char Mensaje, int Envidado, int Pos )
{
 int j ;
 char MiBuffer[20] ;

 sprintf( MiBuffer, "E%1d%c%2d%2d%2d", Quien, Mensaje, Envidado,Partidas[Pos].PuntosActual[0],
	   Partidas[Pos].PuntosActual[1] ) ;
 for( j = 0 ; j<4 ; j++ )
  Envia( Partidas[Pos].Canutos[j], MiBuffer ) ;
}

/* NuevasCartas() Baraja todas las cartas y las reparte al principio de la ronda */
void NuevasCartas( int Pos )
{
 int i,k,temp ;
	/* PREPARAR EL MAZO */
 for( i = 1 ; i <= 40 ; i++ ) Partidas[Pos].Mazo[i-1] = i ;
 for( i = 1 ; i <= 40 ; i++ ) 
  {
   k = random() % 40 ;
   temp = Partidas[Pos].Mazo[k] ;
   Partidas[Pos].Mazo[k]   = Partidas[Pos].Mazo[i-1] ;
   Partidas[Pos].Mazo[i-1] = temp ;
  }
 for( i = 0 ; i < 4 ; i++ )
  {
   Partidas[Pos].Cartas[i][0] = Partidas[Pos].Mazo[i*4+0] ;
   Partidas[Pos].Cartas[i][1] = Partidas[Pos].Mazo[i*4+1] ;
   Partidas[Pos].Cartas[i][2] = Partidas[Pos].Mazo[i*4+2] ;
   Partidas[Pos].Cartas[i][3] = Partidas[Pos].Mazo[i*4+3] ;
  }
 Partidas[Pos].PorDondeVaElMazo = 16 ;
}

/* TieneUnoPares() Funcion de Ayuda que indica si un jugador lleva pares o no */
int TieneUnoPares( int Pos, int Jugador )
{
 int j, k, u, tCarta[4] ;

 for( k=0 ; k<4 ; k++ )
 {
  tCarta[k] = (Partidas[Pos].Cartas[Jugador][k] % 10) * 10 ;
  if( tCarta[k] == 30 || tCarta[k] == 0 ) tCarta[k] = 100 ; /* SI ES UN 3 ES UN REY */
  if( tCarta[k] == 20 ) tCarta[k] = 10 ; /* SI ES UN 2 ES UN AS */
 }
 for( k=0 ; k < 3 ; k++ )
  for( u=k+1 ; u < 4 ; u++ )
    if( tCarta[k] == tCarta[u] ) /* PUES MIRA, TIENE UNA PAREJA */
	return 1 ; /* Tiene PARES */
 return 0 ;
}

/* ValePares() Funcion de ayuda que calcula cuanto valen los pares de una pareja para el tanteo final */
int ValePares( int Pos, int Pareja )
{
 int j, k, u, tCarta[4],Puntos[2] ;

 Puntos[0] = Puntos[1] = 0 ;
 for( j = 0 ; j < 2 ; j++ )
 {
  for( k=0 ; k<4 ; k++ )
  {
   tCarta[k] = (Partidas[Pos].Cartas[Pareja+(j*2)][k] % 10) * 10 ;
   if( tCarta[k] == 30 || tCarta[k] == 0 ) tCarta[k] = 100 ; /* SI ES UN 3 ES UN REY */
   if( tCarta[k] == 20 ) tCarta[k] = 10 ; /* SI ES UN 2 ES UN AS */
  }
  /* PRIMERO VER SI HAY DUPLES */
  if( tCarta[0] == tCarta[1] && tCarta[1] == tCarta[2] && tCarta[2] == tCarta[3] )
   /* TIENE 4 CARTAS IGUALES */
   Puntos[j] = 3 ; /* 3 DE DUPLES */
  else
   if( (tCarta[0] == tCarta[1] && tCarta[2] == tCarta[3]) ||
    (tCarta[0] == tCarta[2] && tCarta[1] == tCarta[3]) ||
    (tCarta[0] == tCarta[3] && tCarta[1] == tCarta[2]) )
     /* TIENE DOBLES PAREJAS */
     Puntos[j] = 3 ; /* 3 DE DUPLES */
   else  /* VEAMOS SI TIENE MEDIAS (= 3 IGUALES) */
    if( tCarta[0] == tCarta[1] && tCarta[1] == tCarta[2] )
     Puntos[j] = 2 ; /* 2 DE MEDIAS */
    else
     if( tCarta[0] == tCarta[1] && tCarta[1] == tCarta[3]  )
      Puntos[j] = 2 ; /* 2 DE MEDIAS */
     else
      if( tCarta[2] == tCarta[3] && tCarta[1] == tCarta[2]  )
	Puntos[j] = 2 ; /* 2 DE MEDIAS */
      else
	if( tCarta[2] == tCarta[3] && tCarta[0] == tCarta[2] )
	 Puntos[j] = 2 ; /* 2 DE MEDIAS */
	else /* VER FINALMENTE SI TIENE PARES */
	 for( k=0 ; k < 3 ; k++ )
	  for( u=k+1 ; u < 4 ; u++ )
	    if( tCarta[k] == tCarta[u] ) /* PUES MIRA, TIENE UNA PAREJA */
		Puntos[j] = 1 ; /* 2 DE PARES */
 }
 return Puntos[0] + Puntos[1] ;
}

/* TieneUnoJuego() Funcion de Ayuda que indica si un jugador lleva Juego o no */
int TieneUnoJuego( int Pos, int Jugador )
{
 int j, k, u, tCarta[4],Puntos ;
 Puntos = 0 ;
 for( k=0 ; k<4 ; k++ )
 {
   switch( Partidas[Pos].Cartas[Jugador][k] % 10 )
   {
    case 8 :
    case 9 :
    case 0 :
    case 3 : Puntos += 10 ; break ;
    case 7 : Puntos += 7  ; break ;
    case 6 : Puntos += 6  ; break ;
    case 5 : Puntos += 5  ; break ;
    case 4 : Puntos += 4  ; break ;
    case 1 :
    case 2 : Puntos += 1  ; break ;
   }
 }
 return (Puntos > 30 ? 1 : 0) ;
}


/* ValeJuego() Funcion de ayuda que calcula cuanto vale las 31 o Punto de una pareja para el tanteo final */
int ValeJuego( int Pos, int Pareja )
{
 int j,k, tCarta[4],Puntos[2] ;

 Puntos[0] = Puntos[1] = 0 ;
 for( j = 0 ; j < 2 ; j++ )
 {
  for( k = 0 ; k<4 ; k++ )
  { /* SUMO PUNTOS POR CARTA */
   switch( Partidas[Pos].Cartas[Pareja+(j*2)][k] % 10 )
   {
    case 8 :
    case 9 :
    case 0 :
    case 3 : Puntos[j] += 10 ; break ;
    case 7 : Puntos[j] += 7  ; break ;
    case 6 : Puntos[j] += 6  ; break ;
    case 5 : Puntos[j] += 5  ; break ;
    case 4 : Puntos[j] += 4  ; break ;
    case 1 :
    case 2 : Puntos[j] += 1  ; break ;
   }
  }
  switch( Puntos[j] )
  {
   case 31: Puntos[j] = Puntos[j] = 3; break ; /* 3 DE 31 */
   case 32:
   case 40:
   case 39:
   case 38:
   case 37:
   case 36:
   case 35:
   case 34:
   case 33: Puntos[j] = 2  ; break ; /* 2 DE JUEGO */
   default: Puntos[j] = 0 ; break ;  /* VETE AL PEO, TOLOMEO, NO TIENES JUEGO */
  }
 }
 if( Puntos[0] + Puntos[1] == 0 ) /* SI NO HAY JUEGO, PUNTO VALE 1 */
  return 1 ;
 return Puntos[0] + Puntos[1] ;
}

/* FinJuego() Algun equipo sumo los puntos suficientes o gano ordago. Fin de juego */
void FinJuego( int Pos )
{
 int j ;

 /* VER SI SE ACABO LA PARTIDA */

 if( Partidas[Pos].VacasActual[0] == Partidas[Pos].CuantasVacas ||
     Partidas[Pos].VacasActual[1] == Partidas[Pos].CuantasVacas )
  {    /* PUES SI, UNA PAREJA ALCANZO LAS VACAS INDICADAS,ENVIAR LA K Y BORRAR ENTRADA DE PARTIDA */
   printf( "Partida concluida: %d\n",Pos ) ;
   sprintf( Buffer2,"K " ) ;
   Buffer2[1] = (Partidas[Pos].VacasActual[0] == Partidas[Pos].CuantasVacas ? '1' : '2' ) ;
   for( j=0 ; j<4 ; j++ )
     Envia( Partidas[Pos].Canutos[j], Buffer2 ) ; /* ADIOS Y GRACIAS POR JUGAR */
   EliminaPartida( Pos ) ;
   return ;
  }
 /* INICIALIZAR VARIABLES, BARAJAR CARTAS, ENVIAR A TODOS LA I Y LA PRIMERA Q */
 Partidas[Pos].PuntosActual[0] = Partidas[Pos].PuntosActual[1] = 0 ;
 if( Partidas[Pos].JuegosActual[0] == Partidas[Pos].JuegosPorVaca || 
     Partidas[Pos].JuegosActual[0] == Partidas[Pos].JuegosPorVaca ) 
    Partidas[Pos].JuegosActual[0] = Partidas[Pos].JuegosActual[1] = 0 ;
 Partidas[Pos].BazaActual= PREGUNTOMUS ;
 Partidas[Pos].Envidado = Partidas[Pos].HayOrdago = Partidas[Pos].PorSumar = 0 ;
 Partidas[Pos].Mano = (Partidas[Pos].Mano + 1) % 4 ;
 Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;
 Partidas[Pos].QuienEnvido = 0 ;
 for( j = 0 ; j<8 ; j++ ) Partidas[Pos].EnvidadoFinal[j] = 0 ;
 Partidas[Pos].QuienGanoPares = Partidas[Pos].QuienGanoJuego = 0 ;
 NuevasCartas( Pos ) ;
 /* REPARTIR LAS CARTAS */
 for( j = 0 ; j < 4 ; j++ )
 {
  sprintf( Buffer, "I%1d%1d%1d%1d%2d%2d%1d%2d%2d%2d%2d",Partidas[Pos].VacasActual[0],Partidas[Pos].VacasActual[1],
	Partidas[Pos].JuegosActual[0],Partidas[Pos].JuegosActual[1],Partidas[Pos].PuntosActual[0],
	Partidas[Pos].PuntosActual[1],Partidas[Pos].Mano,Partidas[Pos].Cartas[j][0],Partidas[Pos].Cartas[j][1],
	Partidas[Pos].Cartas[j][2],Partidas[Pos].Cartas[j][3] ) ;
  Envia( Partidas[Pos].Canutos[j], Buffer ) ;
 }

 /* ENVIAR AL MANO LA PREGUNTA DE SI QUIERE MUS Y EL JUEGO VUELVE A EMPEZAR */
 Buffer[0]='Q' ; Buffer[1] = 0 ;
 Envia( Partidas[Pos].Canutos[Partidas[Pos].Mano], Buffer ) ;
}

/* FinRonda() Calcula los puntos a sumar tras acabar todas las bazas e informa a los jugadores.
   Tambien da inicio a la nueva Ronda o vaca o finaliza la partida si procede */
void FinRonda( int Pos )
{
 int j,k,gGrande,gChica,gPares,gJuego ;
 int     pGrande,pChica,pPares,pJuego ;
 int     ggJuego,ggVaca ;

 /* CALCULAR LOS PUNTOS A SUMAR Y ELABORAR EL MENSAJE P */

 gGrande = gChica = gPares = gJuego = 0 ;
 pGrande = pChica = pPares = pJuego = 0 ;
 ggJuego = ggVaca = 0 ;

 if( Partidas[Pos].EnvidadoFinal[GRANDE] > 0 )
  { /* VEAMOS SI LA GRANDE ESTABA ENVIDADA O EN PASO Y QUIEN SE LO LLEVA */
   gGrande = QuienGanaBaza( Pos,GRANDE ) ;
   if( gGrande > 0 )
    {
     pGrande = Partidas[Pos].EnvidadoFinal[GRANDE] ;
     Partidas[Pos].PuntosActual[gGrande-1] += pGrande ;
    }
  }
  if( Partidas[Pos].PuntosActual[0] < Partidas[Pos].PuntosPorJuego &&
      Partidas[Pos].PuntosActual[1] < Partidas[Pos].PuntosPorJuego )
  { /* VEAMOS SI LA CHICA ESTABA ENVIDADA O EN PASO Y QUIEN SE LO LLEVA */
   if( Partidas[Pos].EnvidadoFinal[CHICA] > 0 )
   {
    gChica = QuienGanaBaza( Pos,CHICA ) ;
    if( gChica > 0 )
     {
      pChica = Partidas[Pos].EnvidadoFinal[CHICA] ;
      Partidas[Pos].PuntosActual[gChica-1] += pChica ;
     }
   }
   if( Partidas[Pos].PuntosActual[0] < Partidas[Pos].PuntosPorJuego &&
      Partidas[Pos].PuntosActual[1] < Partidas[Pos].PuntosPorJuego )
   { /* VEAMOS SI ALGUIEN SE LLEVA LOS PARES Y CUANTO SE LLEVA */
    gPares = Partidas[Pos].QuienGanoPares > 0 ? Partidas[Pos].QuienGanoPares : QuienGanaBaza( Pos,PARES ) ;
    if( gPares > 0 )
    {
     pPares = Partidas[Pos].EnvidadoFinal[PARES] + ValePares( Pos, gPares-1 ) ;
     Partidas[Pos].PuntosActual[gPares-1] += pPares ;
    }
    if( Partidas[Pos].PuntosActual[0] < Partidas[Pos].PuntosPorJuego &&
      Partidas[Pos].PuntosActual[1] < Partidas[Pos].PuntosPorJuego )
    { 
     gJuego = Partidas[Pos].QuienGanoJuego > 0 ? Partidas[Pos].QuienGanoJuego : QuienGanaBaza( Pos,JUEGO ) ;
     if( gJuego == 0 ) gJuego = QuienGanaBaza( Pos,PUNTO ) ;
     pJuego = Partidas[Pos].EnvidadoFinal[JUEGO]+ Partidas[Pos].EnvidadoFinal[PUNTO]+ ValeJuego( Pos, gJuego-1 ) ;
    }
   }
  }
 /* VER SI HA GANADO EL JUEGO Y LA VACA ALGUNA PAREJA */
 if( Partidas[Pos].PuntosActual[0] >= Partidas[Pos].PuntosPorJuego )
  {
   ggJuego = 1 ;
   Partidas[Pos].JuegosActual[0]++ ;
   if(Partidas[Pos].JuegosActual[0] >= Partidas[Pos].JuegosPorVaca )
   {
    ggVaca = 1 ;
    Partidas[Pos].VacasActual[0]++ ;
   }
  }
 else
  if( Partidas[Pos].PuntosActual[1] >= Partidas[Pos].PuntosPorJuego )
   {
    ggJuego = 2 ;
    Partidas[Pos].JuegosActual[1]++ ;
    if(Partidas[Pos].JuegosActual[1] >= Partidas[Pos].JuegosPorVaca )
    {
     ggVaca = 2 ;
     Partidas[Pos].VacasActual[1]++ ;
    }
   }
 /* ENVIAR LA P A TODOS */
 sprintf( Buffer, "P%2d%2d%2d%2d%2d%2d%2d%2d%2d%2d%2d%2d%2d%2d%2d%2d%1d%2d%1d%2d%1d%2d%1d%1d%2d%1d%1d%1d",
  Partidas[Pos].Cartas[0][0],Partidas[Pos].Cartas[0][1],Partidas[Pos].Cartas[0][2],Partidas[Pos].Cartas[0][3],
  Partidas[Pos].Cartas[1][0],Partidas[Pos].Cartas[1][1],Partidas[Pos].Cartas[1][2],Partidas[Pos].Cartas[1][3],
  Partidas[Pos].Cartas[2][0],Partidas[Pos].Cartas[2][1],Partidas[Pos].Cartas[2][2],Partidas[Pos].Cartas[2][3],
  Partidas[Pos].Cartas[3][0],Partidas[Pos].Cartas[3][1],Partidas[Pos].Cartas[3][2],Partidas[Pos].Cartas[3][3],
  gGrande,pGrande,gChica,pChica,gPares,pPares,ValePares(Pos,gPares),gJuego,pJuego,ValeJuego(Pos,gJuego),
  ggJuego,ggVaca ) ;
 for( j = 0 ; j < 4 ; j++ )
  Envia( Partidas[Pos].Canutos[j], Buffer ) ;

 /* VER SI SE HA LLEGADO A 31 (O LO QUE SEA) Y LLAMAR A FINJUEGO() */

 if( Partidas[Pos].PuntosActual[0] >= Partidas[Pos].PuntosPorJuego ||
     Partidas[Pos].PuntosActual[1] >= Partidas[Pos].PuntosPorJuego )
  { FinJuego( Pos ) ; return ; }

 /* INICIAR VARIABLES Y NUEVA RONDA */
 Partidas[Pos].Envidado = Partidas[Pos].HayOrdago = Partidas[Pos].PorSumar = 0 ;
 for( j =0 ; j < 8 ; j++ )
   Partidas[Pos].EnvidadoFinal[j] = 0 ;
 NuevasCartas( Pos ) ;
 Partidas[Pos].Mano = (Partidas[Pos].Mano + 1) % 4 ;
 Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;
 Partidas[Pos].BazaActual = PREGUNTOMUS ;
 Partidas[Pos].QuienGanoPares = Partidas[Pos].QuienGanoJuego = 0 ;
 /* REPARTIR LAS CARTAS */
 for( j = 0 ; j < 4 ; j++ )
 {
  sprintf( Buffer, "I%1d%1d%1d%1d%2d%2d%1d%2d%2d%2d%2d",Partidas[Pos].VacasActual[0],Partidas[Pos].VacasActual[1],
	Partidas[Pos].JuegosActual[0],Partidas[Pos].JuegosActual[1],Partidas[Pos].PuntosActual[0],
	Partidas[Pos].PuntosActual[1],Partidas[Pos].Mano,Partidas[Pos].Cartas[j][0],Partidas[Pos].Cartas[j][1],
	Partidas[Pos].Cartas[j][2],Partidas[Pos].Cartas[j][3] ) ;
  Envia( Partidas[Pos].Canutos[j], Buffer ) ;
 }
 /* ENVIAR AL MANO LA PREGUNTA DE SI QUIERE MUS Y EL JUEGO VUELVE A EMPEZAR */
 Buffer[0]='Q' ; Buffer[1] = 0 ;
 Envia( Partidas[Pos].Canutos[Partidas[Pos].Mano], Buffer ) ;
}

/* TenemosPares() Calcula si en ambos equipos hay al menos un jugador que tenga pares
 o no para pasar a la siguiente baza. Se devuelve true (1) o false (0) */
int TenemosPares( int Pos )
{
 if( ( TieneUnoPares( Pos,0 ) || TieneUnoPares( Pos, 2 ) ) &&
     ( TieneUnoPares( Pos,1 ) || TieneUnoPares( Pos, 3 ) ) )
  return 1 ;
 return 0 ;
}

/* TenemosJuego() Calcula si en ambos equipos hay al menos un jugador por equipo que tenga
 juego para pasar a la siguiente baza. Se devuelve true(1) o false(0) */
int TenemosJuego( int Pos )
{
 if( ( TieneUnoJuego( Pos,0 ) || TieneUnoJuego( Pos, 2 ) ) &&
     ( TieneUnoJuego( Pos,1 ) || TieneUnoJuego( Pos, 3 ) ) )
  return 1 ;
 return 0 ;
}

/* QuienGanaBaza() informa de que pareja tiene mejores cartas en la baza indicada.
   Devuelve 0 si en Pares/Juego nadie tenia. Devuelve 1 si gana la pareja 1
   (Jugadores 0-2). Devuelve 2 si gana la pareja 2 (Jugadores 1-3) */
int QuienGanaBaza( int Pos, int Baza )
{
 int j, k, u, Mayor1, Mayor2, tCarta[4], Sietes, Sotas ;
 long Puntos[4] ;

 Puntos[0] = Puntos[1] = Puntos[2] = Puntos[3] = 0 ;
 switch( Baza )
 {
  case GRANDE :
	for( j = 0 ; j<4 ; j++ )
	{
	 for( k = 0 ; k<4 ; k++ )
	 { /* SUMO PUNTOS POR CARTA */
	  switch( Partidas[Pos].Cartas[j][k] % 10 )
	  {
	   case 1 :
	   case 2 : Puntos[j] += 0 ; break ;	/* ASES Y 2's NO VALEN PARA LA GRANDE */
	   case 4 : Puntos[j] += 5 ; break ;
	   case 5 : Puntos[j] += 25 ; break ;
	   case 6 : Puntos[j] += 125 ; break ;
	   case 7 : Puntos[j] += 625 ; break ;
	   case 8 : Puntos[j] += 3125 ; break ;
	   case 9 : Puntos[j] += 15625 ; break ;
	   case 0 :
	   case 3 : Puntos[j] += 78125; break ; /* AHI ESTA! LOS REYES ES LO QUE MAS VALE */
	  }
	 } /* SUMO PUNTOS POR MANO */
	 if( Partidas[Pos].Mano == j ) Puntos[j] += 3 ;
	  else if( ((Partidas[Pos].Mano + 1) % 4) == j ) Puntos[j] += 2 ;
	   else if( ((Partidas[Pos].Mano +2) % 4 ) == j ) Puntos[j]++ ;
	}
	Mayor1 = Puntos[0] > Puntos[2] ? 0 : 2 ;
	Mayor2 = Puntos[1] > Puntos[3] ? 1 : 3 ;
        if( Puntos[Mayor1] > Puntos[Mayor2] ) return 1 ;
        return 2 ;
	/* NO DEBE HABER EMPATE SI EL ESQUEMA DE PUNTOS ESTA BIEN HECHO!!! */
	break ;

  case CHICA :  /* SI, ALGO SUMA, PERO RECUERDA: JUGADORES DE CHICA = PERDEDORES DE MUS!! */
	for( j = 0 ; j<4 ; j++ )
	{
	 for( k = 0 ; k<4 ; k++ )
	 { /* SUMO PUNTOS POR CARTA */
	  switch( Partidas[Pos].Cartas[j][k] % 10 )
	  {
	   case 0 :
	   case 3 : Puntos[j] += 0 ; break ;	/* Los reyes y los 3's no sirven para chica */
	   case 9 : Puntos[j] += 5 ; break ;
	   case 8 : Puntos[j] += 25 ; break ;
	   case 7 : Puntos[j] += 125 ; break ;
	   case 6 : Puntos[j] += 625 ; break ;
	   case 5 : Puntos[j] += 3125 ; break ;
	   case 4 : Puntos[j] += 15625 ; break ;
	   case 1 :
	   case 2 : Puntos[j] += 78125; break ;  /* Los ases y los 2's es lo que mas vale en chica */
	  }
	 } /* SUMO PUNTOS POR MANO */
	 if( Partidas[Pos].Mano == j ) Puntos[j] += 3 ;
	  else if( ((Partidas[Pos].Mano + 1) % 4) == j ) Puntos[j] += 2 ;
	   else if( ((Partidas[Pos].Mano +2) % 4 ) == j ) Puntos[j]++ ;
	}
	Mayor1 = Puntos[0] > Puntos[2] ? 0 : 2 ;
	Mayor2 = Puntos[1] > Puntos[3] ? 1 : 3 ;
        if( Puntos[Mayor1] > Puntos[Mayor2] ) return 1 ;
        return 2 ;
	/* NO DEBE HABER EMPATE SI EL ESQUEMA DE PUNTOS ESTA BIEN HECHO!!! */
	break ;

  case PARES :
	for( j = 0 ; j<4 ; j++ )
	{
	 for( k=0 ; k<4 ; k++ )
	 {
	  tCarta[k] = (Partidas[Pos].Cartas[j][k] % 10) * 10 ;
	  if( tCarta[k] == 30 || tCarta[k] == 0 ) tCarta[k] = 100 ; /* SI ES UN 3 ES UN REY */
	  if( tCarta[k] == 20 ) tCarta[k] = 10 ; /* SI ES UN 2 ES UN AS */
	 }
	 /* PRIMERO VER SI HAY DUPLES */
	 if( tCarta[0] == tCarta[1] && tCarta[1] == tCarta[2] && tCarta[2] == tCarta[3] )
	   /* TIENE 4 CARTAS IGUALES */
	   Puntos[j] = 5000 + tCarta[0] + tCarta[1] + tCarta[2] + tCarta[3] ;
	 else
	  if( (tCarta[0] == tCarta[1] && tCarta[2] == tCarta[3]) ||
	      (tCarta[0] == tCarta[2] && tCarta[1] == tCarta[3]) ||
	      (tCarta[0] == tCarta[3] && tCarta[1] == tCarta[1]) )
	   /* TIENE DOBLES PAREJAS */
	   Puntos[j] = 5000 + tCarta[0] + tCarta[1] + tCarta[2] + tCarta[3] ;
	   else  /* VEAMOS SI TIENE MEDIAS (= 3 IGUALES) */
	    if( tCarta[0] == tCarta[1] && tCarta[1] == tCarta[2] )
		Puntos[j] = 2000 + tCarta[0] + tCarta[1] + tCarta[2] ;
	    else
	     if( tCarta[0] == tCarta[1] && tCarta[1] == tCarta[3]  )
		Puntos[j] = 2000 + tCarta[0] + tCarta[1] + tCarta[3] ;
	     else
	      if( tCarta[2] == tCarta[3] && tCarta[1] == tCarta[2]  )
		Puntos[j] = 2000 + tCarta[1] + tCarta[2] + tCarta[3] ;
	      else
		if( tCarta[2] == tCarta[3] && tCarta[0] == tCarta[2] )
		 Puntos[j] = 2000 + tCarta[0] + tCarta[2] + tCarta[3] ;
		else /* VER FINALMENTE SI TIENE PARES */
		 for( k=0 ; k < 3 ; k++ )
		  for( u=k+1 ; u < 4 ; u++ )
		    if( tCarta[k] == tCarta[u] ) /* PUES MIRA, TIENE UNA PAREJA */
			Puntos[j] = 1000 + tCarta[k] + tCarta[u] ;
	 /* SUMO PUNTOS POR MANO */
	 if( Partidas[Pos].Mano == j ) Puntos[j] += 3 ;
	  else if( ((Partidas[Pos].Mano + 1) % 4) == j ) Puntos[j] += 2 ;
	   else if( ((Partidas[Pos].Mano +2) % 4 ) == j ) Puntos[j]++ ;
	}
	if( Puntos[0] < 1000 && Puntos[1] < 1000 && Puntos[2] < 1000 && Puntos[3] < 1000 )
		return 0 ; /* RESULTA QUE NINGUNO TENIA PARES */
	Mayor1 = Puntos[0] > Puntos[2] ? 0 : 2 ;
	Mayor2 = Puntos[1] > Puntos[3] ? 1 : 3 ;
        if( Puntos[Mayor1] > Puntos[Mayor2] ) return 1 ;
        return 2 ;
	/* NO DEBE HABER EMPATE SI EL ESQUEMA DE PUNTOS ESTA BIEN HECHO!!! */

	break ;

  case JUEGO :	/* OBJETIVO: SUMAR 31 puntos */
	for( j = 0 ; j<4 ; j++ )
	{
	 for( k = 0 ; k<4 ; k++ )
	 { /* SUMO PUNTOS POR CARTA */
	  switch( Partidas[Pos].Cartas[j][k] % 10 )
	  {
	   case 8 :
	   case 9 :
	   case 0 :
	   case 3 : Puntos[j] += 10 ; break ;
	   case 7 : Puntos[j] += 7  ; break ;
	   case 6 : Puntos[j] += 6  ; break ;
	   case 5 : Puntos[j] += 5  ; break ;
	   case 4 : Puntos[j] += 4  ; break ;
	   case 1 :
	   case 2 : Puntos[j] += 1  ; break ;
	  }
	 }
	switch( Puntos[j] )
	{
	 case 31: Puntos[j] = 100 ; break ; /* 31 ES LO MEJOR QUE SE PUEDE TENER PARA EL JUEGO */
	 case 32: Puntos[j] = 90  ; break ; /* 32 ES LO SEGUNDO MEJOR */
	 case 40: Puntos[j] = 80  ; break ; /* 40 ES LA SIGUIENTE MEJOR PUNTUACION */
	 case 39: Puntos[j] = 70  ; break ;
	 case 38: Puntos[j] = 60  ; break ;
	 case 37: Puntos[j] = 50  ; break ;
	 case 36: Puntos[j] = 40  ; break ;
	 case 35: Puntos[j] = 30  ; break ;
	 case 34: Puntos[j] = 20  ; break ;
	 case 33: Puntos[j] = 10  ; break ;
	 default: Puntos[j] = 0 ; break ;  /* VETE AL PEO, TOLOMEO, NO TIENES JUEGO */
	}

	/* ESTE ES UN BUEN SITIO PARA COMPROBAR SI TIENE LA REAL,
*/
	 Sietes = Sotas = 0 ;
	 for( k = 0 ; k<4 ; k++ ) 
	 { 
	  if( Partidas[Pos].Cartas[j][k] % 10 == 7 ) Sietes++ ;
	  if( Partidas[Pos].Cartas[j][k] % 10 == 8 ) Sotas++ ;
	 }
	 /* AQUI LA REAL SE GANA CON 3 SIETES Y CUALQUIER SOTA */
	 if( Sietes == 3 && Sotas == 1 ) Puntos[j] += 100 ;

	 /* SUMO PUNTOS POR MANO */
	 if( Partidas[Pos].Mano == j ) Puntos[j] += 3 ;
	  else if( ((Partidas[Pos].Mano + 1) % 4) == j ) Puntos[j] += 2 ;
	   else if( ((Partidas[Pos].Mano +2) % 4 ) == j ) Puntos[j]++ ;
	}
	if( Puntos[0] < 10 && Puntos[1] < 10 && Puntos[2] < 10 && Puntos[3] < 10)
		return 0 ; /* RESULTA QUE NINGUNO TENIA JUEGO */
	Mayor1 = Puntos[0] > Puntos[2] ? 0 : 2 ;
	Mayor2 = Puntos[1] > Puntos[3] ? 1 : 3 ;
        if( Puntos[Mayor1] > Puntos[Mayor2] ) return 1 ;
        return 2 ;
	/* NO DEBE HABER EMPATE SI EL ESQUEMA DE PUNTOS ESTA BIEN HECHO!!! */
	break ;

  case PUNTO : /* OBJETIVO: SUMAR 30 PUNTOS PERO NUNCA MAS DE 30*/
	for( j = 0 ; j<4 ; j++ )
	{
	 for( k = 0 ; k<4 ; k++ )
	 { /* SUMO PUNTOS POR CARTA */
	  switch( Partidas[Pos].Cartas[j][k] % 10 )
	  {
	   case 8 :
	   case 9 :
	   case 0 :
	   case 3 : Puntos[j] += 10 ; break ;
	   case 7 : Puntos[j] += 7  ; break ;
	   case 6 : Puntos[j] += 6  ; break ;
	   case 5 : Puntos[j] += 5  ; break ;
	   case 4 : Puntos[j] += 4  ; break ;
	   case 1 :
	   case 2 : Puntos[j] += 1  ; break ;
	  }
	 }
	 if( Puntos[j] > 30 ) Puntos[j] = 0 ; /*ESTO NO DEBE PASAR,PERO POR SI ACASO!!*/
	  else Puntos[j] *= 10 ;
	 /* SUMO PUNTOS POR MANO */
	 if( Partidas[Pos].Mano == j ) Puntos[j] += 3 ;
	  else if( ((Partidas[Pos].Mano + 1) % 4) == j ) Puntos[j] += 2 ;
	   else if( ((Partidas[Pos].Mano +2) % 4 ) == j ) Puntos[j]++ ;
	}
	Mayor1 = Puntos[0] > Puntos[2] ? 0 : 2 ;
	Mayor2 = Puntos[1] > Puntos[3] ? 1 : 3 ;
        if( Puntos[Mayor1] > Puntos[Mayor2] ) return 1 ;
        return 2 ;
	/* NO DEBE HABER EMPATE SI EL ESQUEMA DE PUNTOS ESTA BIEN HECHO!!! */
	break ;
 }
}

/* SiguienteBaza() - Funcion de apoyo que Pasa a la siguiente baza en una partida y envia el 
   mensaje de PARES SI/NO o JUEGO SI/NO a todos */
/* ESTA FUNCION TIENEN UNA PEQUEA TRAMPA: SIMULA QUE LOS USUARIOS MANDAN UN MENSAJE DE CHAT PARA
   INFORMAR DE QUIEN TIENE PARES Y JUEGO EN ESTAS BAZAS, AUNQUE LOS JUGADORES NO MANDEN NINGUN CHAT */
void SiguienteBaza( int Pos )
{
 int j ;
 char cTem[100] ;

	/* PASAR A BAZA SIGUIENTE */
 Partidas[Pos].BazaActual++ ;

 Partidas[Pos].Envidado = 0 ;
 Partidas[Pos].PorSumar = 0 ;
 Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;

	/* SI PARES O JUEGO INDICAR A TODOS QUIEN LO LLEVA */
 if( Partidas[Pos].BazaActual == PARES )
  {
   for ( j = 0 ; j < 4 ; j++ )
     {
      sprintf( cTem,"L%1d%s",j, (TieneUnoPares( Pos, j ) ? "Pares SI" : "Pares NO") ) ;
      Envia( Partidas[Pos].Canutos[0], cTem ) ;
      Envia( Partidas[Pos].Canutos[1], cTem ) ;
      Envia( Partidas[Pos].Canutos[2], cTem ) ;
      Envia( Partidas[Pos].Canutos[3], cTem ) ;
     }
  }
  else
   if( Partidas[Pos].BazaActual == JUEGO )
   {
    for ( j = 0 ; j < 4 ; j++ )
     {
      sprintf( cTem,"L%1d%s",j, (TieneUnoJuego( Pos, j ) ? "Juego SI" : "Juego NO") ) ;
      Envia( Partidas[Pos].Canutos[0], cTem ) ;
      Envia( Partidas[Pos].Canutos[1], cTem ) ;
      Envia( Partidas[Pos].Canutos[2], cTem ) ;
      Envia( Partidas[Pos].Canutos[3], cTem ) ;
     }
   }
	/* SI ESTAMOS EN PARES HAY QUE VER SI TIENEN LAS DOS PAREJAS */
 if( Partidas[Pos].BazaActual == PARES )
  if( !TenemosPares( Pos ) )
    SiguienteBaza( Pos ) ; /* OH Recursividad. Que bien! */
  else  /* VER QUIEN HABLA PRIMERO */
   {
    if( TieneUnoPares( Pos, Partidas[Pos].Mano ) )
	Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;
    else
      if( TieneUnoPares( Pos, (Partidas[Pos].Mano + 1) % 4 ))
	Partidas[Pos].QuienHabla = (Partidas[Pos].Mano + 1) % 4 ;
      else
        Partidas[Pos].QuienHabla = (Partidas[Pos].Mano + 2) % 4 ;
   }

	/* SI ESTAMOS EN JUEGO HAY QUE VER SI TIENEN LAS DOS PAREJAS */
 if( Partidas[Pos].BazaActual == JUEGO )
  if( !TenemosJuego( Pos ) )
    SiguienteBaza( Pos ) ; /* OH Recursividad. Que bien! */
  else  /* VER QUIEN HABLA PRIMERO */
   {
    if( TieneUnoJuego( Pos, Partidas[Pos].Mano ) )
	Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;
    else
      if( TieneUnoJuego( Pos, (Partidas[Pos].Mano + 1) % 4 ))
	Partidas[Pos].QuienHabla = (Partidas[Pos].Mano + 1) % 4 ;
      else
        Partidas[Pos].QuienHabla = (Partidas[Pos].Mano + 2) % 4 ;
   }

	/* SI HEMOS LLEGADO A PUNTO, VER QUE NO HABIA NADIE CON JUEGO */
 if( Partidas[Pos].BazaActual == PUNTO )
   if( TieneUnoJuego( Pos, 0 ) || TieneUnoJuego( Pos, 1 ) || 
       TieneUnoJuego( Pos, 2 ) || TieneUnoJuego( Pos, 3 ) )
	SiguienteBaza( Pos ) ;

	/* VAMOS A VER QUIEN ES EL PRIMERO QUE HABLA EN LA SIGUIENTE BAZA */
 Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;

 if( Partidas[Pos].BazaActual == PARES )
  if( !TieneUnoPares( Pos, Partidas[Pos].Mano ) )
   {
    Partidas[Pos].QuienHabla = (Partidas[Pos].Mano + 1) % 4;
    if( !TieneUnoPares( Pos, (Partidas[Pos].Mano + 1) % 4 ) )
     Partidas[Pos].QuienHabla = (Partidas[Pos].Mano + 2) % 4; /* ESTE YA TIENE PARES SEGURO */
   }

 if( Partidas[Pos].BazaActual == JUEGO )
  if( !TieneUnoJuego( Pos, Partidas[Pos].Mano ) )
   {
    Partidas[Pos].QuienHabla = (Partidas[Pos].Mano + 1) % 4;
    if( !TieneUnoJuego( Pos, (Partidas[Pos].Mano + 1) % 4 ) )
     Partidas[Pos].QuienHabla = (Partidas[Pos].Mano + 2) % 4; /* ESTE YA TIENE JUEGO SEGURO */
   }

	/* VER SI HEMOS ACABADO LA RONDA */
 if( Partidas[Pos].BazaActual > PUNTO )
   FinRonda( Pos ) ;
}

  /* FUNCION IMPORTANTE */
  /* AQUI SE TRATAN LOS MENSAJES RECIBIDOS DE LOS CLIENTES */
void TrataMensaje( int Pos, int Quien, char *Mensaje )
{
 int j,k,Temp,s,CartaTira,Salir ;
 char MiBuffer[100] ;
 /* PRIMERO CORREGIR LA HORA DE LA ULTIMA ACCION */
 Partidas[Pos].UltimaAccion = time( NULL ) ;
 /* Y AHORA VER QUE ES LO QUE QUIERE */
 switch( Mensaje[0] )
 {
  case 'G' : /* UN JUGADOR HACE UNA SEA. Enviar al resto. Es problema suyo como la traten. */
             sprintf( Buffer2, "J%1d%c",Quien, Partidas[Pos].Jugadores[Quien],Mensaje[1] ) ;
             for( j=0 ; j<4 ; j++ )
              if( Partidas[Pos].Canutos[j] > -1 )
               Envia( Partidas[Pos].Canutos[j], Buffer2 ) ;
             return ;
  case 'F' : /* COBARDE!!! ABANDONAS LA PARTIDA!!! */
             sprintf( Buffer2, "BJugador %d [%s] Abandona. La partida finaliza.",Quien, Partidas[Pos].Jugadores[Quien] ) ;
             for( j=0 ; j<4 ; j++ )
              if( Partidas[Pos].Canutos[j] > -1 )
               Envia( Partidas[Pos].Canutos[j], Buffer2 ) ;
             EliminaPartida( Pos ) ;
             return ;
  case 'M' : /* MENSAJE DE CHAT. DECIRSELO A TODOS */
             memcpy( Buffer2, Mensaje+1, 64 ) ;
             Buffer2[64] = 0 ;
             sprintf( Buffer, "L%1d%s",Quien, Buffer2 ) ;
             for( j=0 ; j<4 ; j++ )
              if( Partidas[Pos].Canutos[j] > -1 )
               Envia( Partidas[Pos].Canutos[j], Buffer ) ;
             break ;
  case 'T' :	/* ME DICEN SI QUIEREN MUS O NO */
                if( Partidas[Pos].BazaActual == PREGUNTOMUS && Quien == Partidas[Pos].QuienHabla )
		{
		 sprintf( MiBuffer, "X%1d%c",Quien, Mensaje[1] ) ;
		 for( j=0 ; j<4 ; j++ )
		   Envia( Partidas[Pos].Canutos[j], MiBuffer ) ;
		 if( Mensaje[1] == '1' ) /* SI QUE QUIERE MUS */
		  {
		   Partidas[Pos].QuienHabla= (Quien+1) % 4 ;
		   if( Partidas[Pos].QuienHabla == Partidas[Pos].Mano )/*TODOS SON MUS.EL MANO SE DESCARTA*/
		   {
		    Partidas[Pos].BazaActual = DESCARTE ;
		    Envia( Partidas[Pos].Canutos[Partidas[Pos].QuienHabla], "D" ) ;
		   }
		   else  /* SIGUIENTE HABLA SOBRE SI QUIERE MUS */
		    Envia( Partidas[Pos].Canutos[Partidas[Pos].QuienHabla], "Q" ) ;
		  }
		  else /* NO QUIERE MUS, HABLA MANO A GRANDE */
		  {
		   Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;
		   Partidas[Pos].BazaActual = GRANDE ;
		   Partidas[Pos].Envidado = Partidas[Pos].PorSumar = Partidas[Pos].HayOrdago = 0 ;
		   PreguntaAccion( Pos ) ;
		  }
		}
		break ;
  case 'U' :	/* ME DICEN QUE CARTAS SE QUITAN EN EL DESCARTE */
                if( Partidas[Pos].BazaActual == DESCARTE && Quien == Partidas[Pos].QuienHabla )
		{ /* PRIMERO COMPRUEBO QUE LAS CARTAS QUE ME DICE LAS TIENE */
		 if( (Mensaje[1]-'0') <= 0 || (Mensaje[1]-'0') > 4 )  // SOLO DESCARTE ENTRE 1 Y 4 CARTAS
		  { Envia( Partidas[Pos].Canutos[Quien], "D" ) ; return ; }
		 for( j=0 ; j<(Mensaje[1]-'0') ; j++ )
		 {
		  CartaTira=((Mensaje[(j*2)+2]-'0') * 10) + (Mensaje[(j*2)+3]-'0') ;
		  if( CartaTira < 0 ) CartaTira = Mensaje[(j*2)+3]-'0' ;
		  if( Partidas[Pos].Cartas[Quien][0] == CartaTira ) s = 0 ;
		   else
		    if( Partidas[Pos].Cartas[Quien][1] == CartaTira ) s = 1 ;
		     else
		      if( Partidas[Pos].Cartas[Quien][2] == CartaTira ) s = 2 ;
		       else
		        if( Partidas[Pos].Cartas[Quien][3] == CartaTira ) s = 3 ;
		         else /* ME DICE QUE TIRA UNA CARTA QUE NO TIENE. SO LISTO!! */
			  { Envia( Partidas[Pos].Canutos[Quien], "D" ) ; return ; }
		 }
		 /* ANTES YA HE COMPROBADO QUE SON SUYAS, AHORA LAS CAMBIO */
		 for( j=0 ; j<(Mensaje[1]-'0') ; j++ )
		 {
		  CartaTira=((Mensaje[(j*2)+2]-'0') * 10) + (Mensaje[(j*2)+3]-'0') ;
		  if( CartaTira < 0 ) CartaTira = Mensaje[(j*2)+3]-'0' ;
		  if( Partidas[Pos].Cartas[Quien][0] == CartaTira ) s = 0 ;
		   else
		    if( Partidas[Pos].Cartas[Quien][1] == CartaTira ) s = 1 ;
		     else
		      if( Partidas[Pos].Cartas[Quien][2] == CartaTira ) s = 2 ;
		       else
		        if( Partidas[Pos].Cartas[Quien][3] == CartaTira ) s = 3 ;
		  Temp = Partidas[Pos].Cartas[Quien][s] ;
		  Partidas[Pos].Cartas[Quien][s] = Partidas[Pos].Mazo[Partidas[Pos].PorDondeVaElMazo];
		  Partidas[Pos].Mazo[Partidas[Pos].PorDondeVaElMazo] = Temp ;
		  Partidas[Pos].PorDondeVaElMazo++ ;
		  if( Partidas[Pos].PorDondeVaElMazo == 40 )
		   {
		     Partidas[Pos].PorDondeVaElMazo = 16 ; 
		     for( s=16 ; s<40 ;s++ )
			{
			 k = (random() % 24) + 16 ;
			 Temp = Partidas[Pos].Mazo[s] ;
			 Partidas[Pos].Mazo[s] = Partidas[Pos].Mazo[k] ;
			 Partidas[Pos].Mazo[k] = Temp ; 
			}
		   }
		 }
		 sprintf( MiBuffer, "V%1d%c%2d",Quien, Mensaje[1],40-Partidas[Pos].PorDondeVaElMazo ) ;
		 for( j=0 ; j<4 ; j++ )
		   Envia( Partidas[Pos].Canutos[j], MiBuffer ) ;
		 sprintf( MiBuffer,"W%2d%2d%2d%2d",Partidas[Pos].Cartas[Quien][0],Partidas[Pos].Cartas[Quien][1],
			Partidas[Pos].Cartas[Quien][2],Partidas[Pos].Cartas[Quien][3] ) ;
		 Envia( Partidas[Pos].Canutos[Quien], MiBuffer ) ;
		 Partidas[Pos].QuienHabla = ( Partidas[Pos].QuienHabla + 1 ) % 4 ;
		 if( Partidas[Pos].QuienHabla == Partidas[Pos].Mano )
		 {            /* PREGUNTO MUS OTRA VEZ */
		  Partidas[Pos].BazaActual = PREGUNTOMUS ;
		  Envia( Partidas[Pos].Canutos[Partidas[Pos].QuienHabla], "Q" ) ;
		 }
		 else         /* SE DESCARTA EL SIGUIENTE */
		   Envia( Partidas[Pos].Canutos[Partidas[Pos].QuienHabla], "D" ) ;
		}
		break ;

  case 'H' :	/* IMPORTANTE MENSAJE!!! ME DICEN QUE ACCION: ENVIDO, PASO, SUBO, ORDAGO O VEO */
		if( Quien != Partidas[Pos].QuienHabla || Partidas[Pos].BazaActual < GRANDE )
		  return ;					  /* ENVIDA PERO ...*/
		if( Mensaje[1] == 'E' && Partidas[Pos].PorSumar ) /* YA ESTABA ENVIDADO */
		 { PreguntaAccion( Pos ) ; return ; }                 /* SUBE PERO ...*/
		if( Mensaje[1] == 'S' && Partidas[Pos].PorSumar == 0) /* NADIE ENVIDO TODAVIA */
		 { PreguntaAccion( Pos ) ; return ; }                  /* LO VE PERO ...*/
		if( Mensaje[1] == 'V' && Partidas[Pos].PorSumar == 0 ) /* NO HAY NADA QUE VER */
		 { PreguntaAccion( Pos ) ; return ; }              /* ECHA ORDAGO PERO ...*/
		if( Mensaje[1] == 'O' && Partidas[Pos].HayOrdago ) /* YA HABIA UN ORDAGO ECHADO*/
		 { PreguntaAccion( Pos ) ; return ; }

		/* BIEN, EN ESTE PUNTO EL JUGADOR HA HECHO UNA ACCION LEGAL, VEAMOS EL RESULTADO */
		switch( Mensaje[1] )
		{
		 case 'P' :	/* JUGADOR PASA */
			if( Partidas[Pos].Envidado == 0 )           /* SI LA BAZA IBA EN PASO */
			{  /* QUE HABLE EL SIGUIENTE */
			 Partidas[Pos].QuienHabla = (Partidas[Pos].QuienHabla + 1) % 4 ;
			 MensajeTrasAccion( Quien, Mensaje[1],0, Pos ) ; /* INFORMAR DE LO QUE HA HECHO */
			   /* SI ES PARES, VER SI AL QUE LE TOCA LOS LLEVA. SI NO, PASAR AL SIGUIENTE */
			 if( Partidas[Pos].BazaActual == PARES && !TieneUnoPares( Pos,Partidas[Pos].QuienHabla )
			     && Partidas[Pos].QuienHabla != Partidas[Pos].Mano	)
			 {
			  Partidas[Pos].QuienHabla = (Partidas[Pos].QuienHabla + 1) % 4 ;
			  if( Partidas[Pos].BazaActual == PARES && !TieneUnoPares( Pos,Partidas[Pos].QuienHabla )
			     && Partidas[Pos].QuienHabla != Partidas[Pos].Mano	)
			    Partidas[Pos].QuienHabla = (Partidas[Pos].QuienHabla + 1) % 4 ;
			 }
			   /* SI ES JUEGO, VER SI AL QUE LE TOCA LO LLEVA. SI NO, PASAR AL SIGUIENTE */
			 if( Partidas[Pos].BazaActual == JUEGO && !TieneUnoJuego( Pos,Partidas[Pos].QuienHabla )
			     && Partidas[Pos].QuienHabla != Partidas[Pos].Mano	)
			 {
			  Partidas[Pos].QuienHabla = (Partidas[Pos].QuienHabla + 1) % 4 ;
			  if( Partidas[Pos].BazaActual == JUEGO && !TieneUnoJuego( Pos,Partidas[Pos].QuienHabla )
			     && Partidas[Pos].QuienHabla != Partidas[Pos].Mano	)
			    Partidas[Pos].QuienHabla = (Partidas[Pos].QuienHabla + 1) % 4 ;
			 }

			 if( Partidas[Pos].QuienHabla == Partidas[Pos].Mano ) /* BAZA EN PASO */
			  if( Partidas[Pos].BazaActual >= JUEGO ) /* SI ERA JUEGO o PUNTO, SE ACABO LA RONDA. */
			   { FinRonda( Pos ) ; return ; }
			  else /* SE VE UNA PARA EL FINAL Y SE PASA DE BAZA */
			   {
			    if(Partidas[Pos].BazaActual <= CHICA ) 
			     Partidas[Pos].EnvidadoFinal[Partidas[Pos].BazaActual] = 1 ;
			    SiguienteBaza( Pos ) ;
			   }
			 if( Partidas[Pos].BazaActual <= PUNTO )
			  PreguntaAccion( Pos ) ;
			 else
			  return ;
			}
			else                          /* LA BAZA IBA ENVIDADA */
			{
			 Partidas[Pos].QuienHabla = (Partidas[Pos].QuienHabla + 2) % 4 ;
			  /* VER SI YA HAN HABLADO LOS DOS DE LA PAREJA ENVIDADA O EL NUEVO JUGADOR
			     NO LLEVA PARES O JUEGO EN ESA BAZA */
			/* CURIOSA FORMULA PARA VER QUIEN HABLABA PRIMERO TRAS UN ENVITE */
			 Temp = (Partidas[Pos].QuienEnvido + Partidas[Pos].Mano) % 2 ? Partidas[Pos].Mano : Partidas[Pos].Mano + 1 ;
			 if((Partidas[Pos].QuienHabla == Temp ) ||
			    (Partidas[Pos].BazaActual == PARES && !TieneUnoPares(Pos,Partidas[Pos].QuienHabla )) ||
			    (Partidas[Pos].BazaActual == JUEGO && !TieneUnoJuego(Pos,Partidas[Pos].QuienHabla )) )
			 {      /* NO ACEPTARON EL ENVITE U ORDAGO ASI QUE SE LO APUNTAN LOS OTROS */
			  Partidas[Pos].PuntosActual[Partidas[Pos].QuienEnvido % 2] += Partidas[Pos].PorSumar ;
				/* CON LO QUE SUMA VER SI HAN GANADO EL JUEGO */
			  MensajeTrasAccion( Quien, Mensaje[1],0, Pos ) ;
			  if( Partidas[Pos].BazaActual == PARES )
			   Partidas[Pos].QuienGanoPares = (Partidas[Pos].QuienEnvido % 2) + 1 ;
			  if( Partidas[Pos].BazaActual >= JUEGO )
			   Partidas[Pos].QuienGanoJuego = (Partidas[Pos].QuienEnvido % 2) + 1 ;
			  if( Partidas[Pos].PuntosActual[Partidas[Pos].QuienEnvido % 2] >= Partidas[Pos].PuntosPorJuego )
				{ FinJuego( Pos ) ; return ; }
			  else
			  {
			   MensajeTrasAccion( Quien, Mensaje[1],0, Pos ) ;
			   Partidas[Pos].Envidado = Partidas[Pos].PorSumar = Partidas[Pos].HayOrdago = 0 ;
			   Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;
			   SiguienteBaza( Pos ) ;   /* PASAR A BAZA SIGUIENTE */
			   if( Partidas[Pos].BazaActual > PUNTO )
			     return ;
			  }
			 }
			 PreguntaAccion( Pos ) ;
			}
			break ;

		case 'E' :	/* JUGADOR ENVIDA */
			MiBuffer[2] = 0 ;
			MiBuffer[0] = Mensaje[2] ;
			MiBuffer[1] = Mensaje[3] ;
			Partidas[Pos].Envidado = atoi( MiBuffer ) ;
			if( Partidas[Pos].Envidado == 0 ) Partidas[Pos].Envidado = 2 ;
			Partidas[Pos].PorSumar = 1 ;
			Partidas[Pos].QuienEnvido = Quien ;
			MensajeTrasAccion( Quien, Mensaje[1],Partidas[Pos].Envidado, Pos ) ;
			Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;
			/* HABLA EL MANO O EL SIGUIENTE AL MANO  SI ES PAREJA DEL QUE HABLO */
			if( (Quien - Partidas[Pos].Mano) % 2 == 0 )
			  Partidas[Pos].QuienHabla = (Partidas[Pos].Mano + 1) % 4 ;
			else
			  Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;
			/* SI EL QUE HABLA NO TIENEPARES O JUEGO EN ESA BAZA, TIENE QUE HABLAR SU COMPAERO */
			if( Partidas[Pos].BazaActual == PARES && !TieneUnoPares(Pos,Partidas[Pos].QuienHabla ) ||
			    Partidas[Pos].BazaActual == JUEGO && !TieneUnoJuego(Pos,Partidas[Pos].QuienHabla ) )
				Partidas[Pos].QuienHabla = (Partidas[Pos].QuienHabla + 2) % 4 ;
			PreguntaAccion( Pos ) ;
			break ;

		case 'S' :	/* JUGADOR SUBE EL ENVITE */
			MiBuffer[2] = 0 ;
			MiBuffer[0] = Mensaje[2] ;
			MiBuffer[1] = Mensaje[3] ;
			Partidas[Pos].PorSumar = Partidas[Pos].Envidado ;
			Partidas[Pos].Envidado = Partidas[Pos].Envidado + (atoi(MiBuffer)?atoi(MiBuffer):1) ;
			Partidas[Pos].QuienEnvido = Quien ;
			MensajeTrasAccion( Quien, Mensaje[1],atoi(MiBuffer), Pos ) ;
			if( (Quien - Partidas[Pos].Mano) % 2 == 0 )
			  Partidas[Pos].QuienHabla = (Partidas[Pos].Mano + 1) % 4 ;
			else
			  Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;
			/* SI EL QUE HABLA NO TIENE PARES O JUEGO EN ESA BAZA, TIENE QUE HABLAR SU COMPAERO */
			if( Partidas[Pos].BazaActual == PARES && !TieneUnoPares(Pos,Partidas[Pos].QuienHabla ) ||
			    Partidas[Pos].BazaActual == JUEGO && !TieneUnoJuego(Pos,Partidas[Pos].QuienHabla ) )
				Partidas[Pos].QuienHabla = (Partidas[Pos].QuienHabla + 2) % 4 ;
			PreguntaAccion( Pos ) ;
			break ;

		case 'O' :	/* JUGADOR METE ORDAGO */
			Partidas[Pos].HayOrdago = 1 ;
			Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;
			Partidas[Pos].QuienEnvido = Quien ;
			Partidas[Pos].PorSumar = Partidas[Pos].Envidado ;
			if( Partidas[Pos].PorSumar == 0 ) Partidas[Pos].PorSumar = 1 ;
			Partidas[Pos].Envidado = Partidas[Pos].PuntosPorJuego ;
			MensajeTrasAccion( Quien, Mensaje[1],0, Pos ) ;
			if( (Quien - Partidas[Pos].Mano) % 2 == 0 )
			  Partidas[Pos].QuienHabla = (Partidas[Pos].Mano + 1) % 4 ;
			else
			  Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;
			PreguntaAccion( Pos ) ;
			/* SI EL QUE HABLA NO TIENEPARES O JUEGO EN ESA BAZA, TIENE QUE HABLAR SU COMPAERO */
			if( Partidas[Pos].BazaActual == PARES && !TieneUnoPares(Pos,Partidas[Pos].QuienHabla ) ||
			    Partidas[Pos].BazaActual == JUEGO && !TieneUnoJuego(Pos,Partidas[Pos].QuienHabla ) )
				Partidas[Pos].QuienHabla = (Partidas[Pos].QuienHabla + 2) % 4 ;
			break ;

		case 'V' :	/* JUGADOR VE EL ORDAGO O ENVITE */
			if( Partidas[Pos].HayOrdago ) /* EL JUGADOR HA VISTO UN ORDAGO */
			{
			 j = QuienGanaBaza( Pos, Partidas[Pos].BazaActual ) ;
			 Partidas[Pos].PuntosActual[j] = Partidas[Pos].PuntosPorJuego ;
			 MensajeTrasAccion( Quien,Mensaje[1],0,Pos ) ;
			 FinRonda( Pos ) ;
			}
			else /* SE VE EL ENVITE, SE APUNTA PARA EL FINAL Y SE PASA DE BAZA */
			{
			 MensajeTrasAccion( Quien,Mensaje[1],0,Pos ) ;
			 Partidas[Pos].EnvidadoFinal[Partidas[Pos].BazaActual] = Partidas[Pos].Envidado ;
			 /* CAMBIAR DE BAZA FINDE JUEGO SI ESTABAMOS EN JUEGO O PUNTO */
			 SiguienteBaza( Pos ) ;
			 if( Partidas[Pos].BazaActual > PUNTO )
			    { FinRonda( Pos ) ; return ; }
			 PreguntaAccion( Pos ) ;
			}
			break ;
		}
		break ;
 }
}

void PonEnListaWeb( int s )
{
 if( CuantosWebPendientes == MAXWEBPEND ) /* LO SIENTO PERO NO HAY SITIO */
  {
   close ( s ) ;
   return ;
  }
 /* ESTE ES UN BUEN MOMENTO PARA HACERLO NO BLOQUEANTE */
 ioctl( s, FIONBIO, &bloquea ) ;

 SocketWebPendientes[ CuantosWebPendientes++ ] = s ;
}

void PonEnListaMus( int s ) 
{
 if( CuantosMusPendientes == MAXMUSPEND ) /* LO SIENTO PERO NO HAY SITIO */
  {
   close ( s ) ;
   return ;
  }
 /* ESTE ES UN BUEN MOMENTO PARA HACERLO NO BLOQUEANTE */
 ioctl( s, FIONBIO, &bloquea ) ;

 SocketMusPendientes[ CuantosMusPendientes++ ] = s ;
}

void SacaDeListaWeb( int Pos )
{
 int j ;
 if( Pos<0 ) return ; /* POR SI LAS FLY's */
 for( j = Pos ; j < CuantosWebPendientes-1 ; j++ )
   SocketWebPendientes[ j ] = SocketWebPendientes[ j+1 ] ;
 CuantosWebPendientes-- ;
}

void SacaDeListaMus( int Pos )
{
 int j ;
 if( Pos<0 ) return ; /* POR SI LAS FLY's */
 for( j = Pos ; j < CuantosMusPendientes-1 ; j++ )
   SocketMusPendientes[ j ] = SocketMusPendientes[ j+1 ] ;
 CuantosMusPendientes-- ;
}

/* Empieza Partida. Inicializar variables y mandar mensajes indicandolo */
void EmpiezaPartida( int Pos )
{
 int i,temp,k ; 

  /** EMPIEZA EL PRIMER JUEGO DE LA PRIMERA VACA **/
  /** BARAJAR, REPARTIR CARTAS **/
/* Inicializar contadores */
 Partidas[Pos].BazaActual = PREGUNTOMUS ;
 Partidas[Pos].PuntosActual[0] = 0 ;
 Partidas[Pos].PuntosActual[1] = 0 ;
 Partidas[Pos].JuegosActual[0] = 0 ;
 Partidas[Pos].JuegosActual[1] = 0 ;
 Partidas[Pos].VacasActual[0] = 0 ;
 Partidas[Pos].VacasActual[1] = 0 ;
 Partidas[Pos].Envidado  = 0 ;
 Partidas[Pos].HayOrdago = 0 ;
 Partidas[Pos].PorSumar  = 0 ;
 for( i = 0 ; i<8 ; i++ ) Partidas[Pos].EnvidadoFinal[i] = 0 ;
 Partidas[Pos].QuienGanoPares = Partidas[Pos].QuienGanoJuego = 0 ;
 		/* AH!! EL MANO SE ELIJE AL AZAR EN LA VERSION 1.0
		   SE PODRA HACER UN CORRIDO Y SIN SEAS EN LA PROXIMA VERSION */
 Partidas[Pos].Mano = random() % 4 ;
 Partidas[Pos].QuienHabla = Partidas[Pos].Mano ;

 /* BARAJAR Y REPARTIR */
 NuevasCartas( Pos ) ;

 /* ENVIAR A LOS 4 JUGADORES LA C */
 for( i = 0 ; i < 4 ; i++ )
 {
  sprintf( Buffer, "C%1d%1d%2d%1d%16s%16s%16s%16s",Partidas[Pos].CuantasVacas,Partidas[Pos].JuegosPorVaca,
          Partidas[Pos].PuntosPorJuego,i,Partidas[Pos].Jugadores[0],Partidas[Pos].Jugadores[1],
          Partidas[Pos].Jugadores[2],Partidas[Pos].Jugadores[3] ) ;
  Envia( Partidas[Pos].Canutos[i], Buffer ) ;
 }

 /* REPARTIR LAS CARTAS */
 for( i = 0 ; i < 4 ; i++ )
 {
  sprintf( Buffer, "I00000000%1d%2d%2d%2d%2d",Partidas[Pos].Mano,Partidas[Pos].Cartas[i][0],
	Partidas[Pos].Cartas[i][1],Partidas[Pos].Cartas[i][2],Partidas[Pos].Cartas[i][3] ) ;
  Envia( Partidas[Pos].Canutos[i], Buffer ) ;
 }

 /* ENVIAR AL MANO LA PREGUNTA DE SI QUIERE MUS Y LA PARTIDA HA EMPEZADO */
 Buffer[0]='Q' ; Buffer[1] = 0 ;
 Envia( Partidas[Pos].Canutos[Partidas[Pos].Mano], Buffer ) ;
}

void RespondeHTTPPendiente( int Pos )
{
 int j ,kk ,faltan ; 
 char tmpCad[100], *p1, *p2 ;
 FILE *f ;

 /* VER EL FICHERO QUE PIDEN */
 Buffer[400] = 0 ;
 p1 = strstr( Buffer, "GET" ) ;
 if( p1 == NULL ) p1 = strstr( Buffer, "Get" ) ;
 if( p1 == NULL ) p1 = strstr( Buffer, "get" ) ;
 if( p1 != NULL )  /* ENCONTRADO EL NOMBRE DE FICHERO */
  {
   j = 5 ;
   kk = 0 ;
   if( j < strlen(p1) ) /* EVITO DESBORDAMIENTO DE BUFFER */
     while ( p1[j] != ' ' && p1[j] )
	tmpCad[kk++] = p1[j++] ;
   tmpCad[kk] = 0 ;
  }
printf( "Sirve Web [%s]\n",tmpCad ) ;
	/* AQUI SE HACE EL SERVICIO DE PAGINAS WEB. OPCION sumus -d !ALTAMENTE DESACONSEJABLE! */
 if( SirveWeb && strcmp( tmpCad,"/" ) && tmpCad[0] ) /* SERVIR LA PAGINA */
 {
  if( strstr( tmpCad, ".jar" ) || strstr( tmpCad, ".class" ) ) p1 = "application/x-java" ;
  else if( strstr( tmpCad, ".jpg" ) || strstr( tmpCad, ".jpeg" ) ) p1 = "image/jpeg" ;
  else if( strstr( tmpCad, ".gif" ) ) p1 = "image/gif" ;
  else if( strstr( tmpCad, ".txt" ) ) p1 = "text/plain" ;
  else p1 = "text/html" ;
  /* ABRIR EL FICHERO PEDIDO EN EL SUBDIRECTORIO ./web */
  sprintf( Buffer, "./web/%s", tmpCad ) ;
  f = fopen( Buffer, "rb" ) ;
  if( f != NULL )
  {
   sprintf( Buffer, "HTTP/1.0 200 OK\nContent-type: %s%c%c", p1,10,10 ) ;
   send( SocketWebPendientes[ Pos ], Buffer, strlen(Buffer), MSG_NOSIGNAL ) ;
   do
    {		/* LEER TODO EL FICHERO Y ENVIARLO */
      j = fread( Buffer, 1, 65535, f ) ;
      if( j ) send( SocketWebPendientes[ Pos ], Buffer, j, MSG_NOSIGNAL ) ;
    } while( j == 65535 ) ;
  }
  else
   {
    strcpy( Buffer, "HTTP/1.0 404 FILE NOT FOUND\nContent-type: text/html\n\n<h1>404 FILE NOT FOUND.</h1>" ) ;
    send( SocketWebPendientes[ Pos ], Buffer, strlen(Buffer), MSG_NOSIGNAL ) ;
   }
  if( f != NULL ) fclose( f ) ;
 }
 else
 {
  send( SocketWebPendientes[ Pos ], CabeceraHTTP, strlen(CabeceraHTTP), MSG_NOSIGNAL ) ;
  sprintf( Buffer,"Puerto por el que escucha este servidor de partidas: %d <br>
                  Numero maximo de partidas simultaneas: %d<br>
                  Numero de partidas disputandose actualmente: %d<br>", PuertoMus, MaxPartidas, NumPartidas ) ;
  strcat( Buffer, "<br>Listado de partidas en curso:<br><UL>" ) ;
  send( SocketWebPendientes[ Pos ], Buffer, strlen(Buffer), MSG_NOSIGNAL ) ;

  for( j = 0 ; j < MaxPartidas ; j++ )
  {
   if( Partidas[j].NombrePartida[0] )
    {
      strcpy( Buffer,"<br><LI><B>Nombre Partida: ") ;
      strcat( Buffer, Partidas[j].NombrePartida ) ;
      faltan = 4 ;
      for( kk = 0 ; kk < 4 ; kk++ )
	if( Partidas[j].Canutos[kk] > -1 ) faltan-- ;
      sprintf( tmpCad, "Faltan %d jugador(es) para completar la partida." , faltan ) ;
      sprintf( Buffer2, "</B></li><LI>Numero de Partida: %d</li><li>Creada por: %s</li>
                        <li>%s es una partida privada.</li><li>%s</LI><br><br>",j,Partidas[j].Jugadores[0],
               Partidas[j].EsPublica ? "NO": "SI",(Partidas[j].Canutos[1]>-1 &&
               Partidas[j].Canutos[2]>-1 && Partidas[j].Canutos[3]>-1 ) ? "Jugadores ya inscritos. Partida en marcha." :
               tmpCad ) ;
      strcat( Buffer,Buffer2 ) ;
      send( SocketWebPendientes[ Pos ], Buffer, strlen(Buffer), MSG_NOSIGNAL ) ;
    }
  }
  send( SocketWebPendientes[ Pos ], FinalHTTP, strlen(FinalHTTP), MSG_NOSIGNAL ) ;
 }
 close( SocketWebPendientes[ Pos ] ) ;
 SacaDeListaWeb( Pos ) ;
}

/* EN ESTA FUNCION SE GESTIONAN LAS PETICIONES DE CREACION DE PARTIDAS DE MUS Y 
   SOLICITUD DE APUNTARSE A ELLAS */

void RespondeMusPendiente( int Pos )
{
 int j,k,conta ;
 char Cad[200] ;
 Mensaje = (struct sMensaje *) Buffer ;
 if( Mensaje->id == 'A' )
 {  /* APUNTARSE A UNA PARTIDA */
  strncpy( Buffer2, Mensaje->a.c_ApunPart.NombrePartida, 16 ) ;
  Buffer2[16]=0 ;
  k = -1 ;
  j = PosPrimeraPartida ;
  while( j >= 0 && k < 0 )
  {
   if( !strncmp( Mensaje->a.c_ApunPart.NombrePartida, Partidas[j].NombrePartida, 16 ) )
     k = j ;
   else
     j = Partidas[j].Siguiente ;
  }
  if( k == -1 )  /* PARTIDA NO ENCONTRADA */
  {
    strcpy( Buffer2, "R0Nombre de partida no encontrado" ) ;
    Envia( SocketMusPendientes[ Pos ], Buffer2 ) ;
    close( SocketMusPendientes[ Pos ] ) ;
    SacaDeListaMus( Pos ) ;
    return ;
  }
  conta = 0 ;
  if( !Partidas[k].EsPublica )
  {		/* ES PARTIDA PRIVADA. VER SI ESTA APUNTADO */
   for( j = 1 ; j < 4 ; j++ )
    if( !strncmp( Partidas[k].Jugadores[j], Mensaje->a.c_ApunPart.NombreJugador, 16 ) )
     { Partidas[k].Canutos[j] = SocketMusPendientes[ Pos ] ; conta = j ; }
   if( conta ) /* Encontrado y Apuntado a la partida */
    {
     strcpy( Buffer2, "R1" ) ;
     Envia( SocketMusPendientes[ Pos ], Buffer2 ) ;     
     SacaDeListaMus( Pos ) ;
     if( Partidas[k].Canutos[1]>=0 && Partidas[k].Canutos[2]>=0 && Partidas[k].Canutos[3]>=0 )
       EmpiezaPartida( k ) ; /*YA SE HAN APUNTADO LOS 4. EMPIEZA PARTIDA. ENVIAR 'C'*/
     return ;
    }
   else /* NO ESTA EN LA LISTA DE JUGADORES ADMITIDOS */
    {
    strcpy( Buffer2, "R0Nombre de jugador no encontrado" ) ;
    Envia( SocketMusPendientes[ Pos ], Buffer2 ) ;
    close( SocketMusPendientes[ Pos ] ) ;
    SacaDeListaMus( Pos ) ;
    return ;
    }
   }
  else  /* ES PARTIDA PUBLICA. SE LE APUNTA SI HAY HUECO */
   {
    conta = 0 ;
    for( j = 1 ; j < 4 ; j++ )
     if( Partidas[k].Canutos[j] == -1 )
       { 
        conta = 1 ; 
        Partidas[k].Canutos[j] = SocketMusPendientes[ Pos ] ;
        strncpy( Partidas[k].Jugadores[j], Mensaje->a.c_ApunPart.NombreJugador,16 ) ;
        Partidas[k].Jugadores[j][16]=0 ;
        j= 4 ;
       }
     if( conta ) /* HABIA HUECO PARA QUE JUGARA */
      {
       strcpy( Buffer2, "R1" ) ;
       Envia( SocketMusPendientes[ Pos ], Buffer2 ) ;     
       SacaDeListaMus( Pos ) ;
       if( Partidas[k].Canutos[0] > -1 && Partidas[k].Canutos[1] > -1 && Partidas[k].Canutos[2] > -1 &&
           Partidas[k].Canutos[3] > -1 ) /*YA SE HAN APUNTADO LOS 4. EMPIEZA PARTIDA. ENVIAR 'C'*/
       EmpiezaPartida( k ) ;
       return ;       
      }
     else
      {
       strcpy( Buffer2, "R0Partida ya completa") ;
       Envia( SocketMusPendientes[ Pos ], Buffer2 ) ;
       close( SocketMusPendientes[ Pos ] ) ;
       SacaDeListaMus( Pos ) ;
       return ;
      }
   }
 }
 else
  if( Mensaje->id == 'N' )
  { /* CREAR PARTIDA NUEVA */
   if( NumPartidas==MaxPartidas )  /* QUEDA HUECO PARA OTRA PARTIDA??? */
   {
    strcpy( Buffer2, "R0No hay sitio para mas partidas  " ) ;
    Envia( SocketMusPendientes[ Pos ], Buffer2 ) ;
    close( SocketMusPendientes[ Pos ] ) ;
    SacaDeListaMus( Pos ) ;
    return ;
   }
   k = 0 ;    /* BUSCAR PRIMER HUECO PARA ESTA PARTIDA */
   while( Partidas[k].NombrePartida[0] && k < MaxPartidas ) k++ ;
   conta = -1 ;
   j = PosPrimeraPartida ;  /* Meter esta partida en la lista, encadenarla */
   while( j > -1 )
     { conta = j ; j = Partidas[j].Siguiente ; }
   if( conta > -1 )
     Partidas[conta].Siguiente = k ;
   else
     PosPrimeraPartida = k ;
   Partidas[k].Siguiente = -1 ;
   Partidas[k].Anterior = conta ;
   Partidas[k].EsPublica= (Mensaje->a.c_CreaPart.Privada == '1' ? 0 : 1 ) ;
   Partidas[k].Canutos[0] = SocketMusPendientes[Pos] ;
   Partidas[k].Canutos[1] = Partidas[k].Canutos[2] = Partidas[k].Canutos[3] = -1 ;
   strncpy( Partidas[k].Jugadores[0], Mensaje->a.c_CreaPart.Jugador,16 ) ;
   Partidas[k].Jugadores[0][16] = 0 ;
   Partidas[k].CuantasVacas = Mensaje->a.c_CreaPart.Vacas - '0' ;
   Partidas[k].JuegosPorVaca = Mensaje->a.c_CreaPart.JuegosPorVaca - '0' ;
   Partidas[k].PuntosPorJuego = (Mensaje->a.c_CreaPart.PuntosPorJuego[0] - '0') * 10 +
                                (Mensaje->a.c_CreaPart.PuntosPorJuego[1] - '0')  ;
   if( !Partidas[k].EsPublica )
   {
    strncpy( Partidas[k].Jugadores[1], Mensaje->a.c_CreaPart.Jugador1,16 ) ;
    strncpy( Partidas[k].Jugadores[2], Mensaje->a.c_CreaPart.Companero,16 ) ;
    strncpy( Partidas[k].Jugadores[3], Mensaje->a.c_CreaPart.Jugador3,16 ) ;
    Partidas[k].Jugadores[1][16] = 0 ;
    Partidas[k].Jugadores[2][16] = 0 ;
    Partidas[k].Jugadores[3][16] = 0 ;
   }
   strncpy( Partidas[k].NombrePartida, Mensaje->a.c_CreaPart.NombrePartida,16 ) ;
   Partidas[k].NombrePartida[16]=0 ;
   Partidas[k].UltimaAccion = time( NULL ) ;/*para que no se quede toda la vida se le pone fecha*/
   strcpy( Buffer2, "R1" ) ;
   Envia( SocketMusPendientes[ Pos ], Buffer2 ) ;     
   SacaDeListaMus( Pos ) ;
   printf( "Creada Partida %s:[%s] en %d por %s\n",(Partidas[k].EsPublica?"Publica":"Privada"),Partidas[k].NombrePartida, k,Partidas[k].Jugadores[0] ) ;
   NumPartidas++ ;
   return ;       
  }
  else
   { /* Este no ha respetado el protocolo. ADIOS! */
     close( SocketMusPendientes[ Pos ] ) ;
     SacaDeListaMus( Pos ) ;
   }
}

/********** Funcion MAIN *************/

int main( int argc, char ** argv )
{
 int j,k,l,p ; /* Famosas variables que sirven para todo */
 int PonAyuda ;
 int NuevoSocket ;
 struct sockaddr_in host ;
 time_t Ahora ;

 /* AQUI EMPIEZA LA EJECUCION DEL PROGRAMA */

 /* INICIALIZAMOS UNAS CUANTAS VARIABLES */ 
 PonAyuda = 0 ;
 MaxPartidas= 5 ;
 PuertoWeb= 81 ;
 PuertoMus= 4321 ;
 k = 0 ;

 /* Como aqui hay barajeo de cartas, tratamos de hacerlo suficientemente aleatorios */
 srandom( time( NULL ) ) ;

 /* LO PRIMERO ES VER LOS PARAMETROS QUE LE PASAN */
 for( j =0 ; j < argc ; j++ )
 {
  if( argv[j][0] == '-' )
  {  
   switch( argv[j][1] )
   {
    case 'w' :	k = atoi( argv[j+1] ) ;	/* INDICAN EL PUERTO QUE ESCUCHA HTTP */
		if( k>0 ) PuertoWeb = k ;
		break ;
    case 'p' :	k = atoi( argv[j+1] ) ; /* INDICAN EL PUERTO QUE ESCUCHA PARTIDAS DE MUS */
		if( k>0 ) PuertoMus = k ;
		break ;
    case 'm' :	k = atoi( argv[j+1] ) ; /* INDICAN EL MAXIMO DE PARTIDAS SIMULTANEAS */
		if( k>0 ) MaxPartidas = k ;
		break ;
    case 'd' :  SirveWeb = 1 ;          /* INDICAN SI SE COMPORTA COMO SERVIDOR WEB DE PAGINAS */
	        break ;
    default:  /* SI PUSO ALGO RARO (-h) QUE LE SAQUE LA AYUDA */
		PonAyuda=1;
   }
  }
 }
 if( PonAyuda )
 {
  printf( "sumus: Servidor de partidas de Mus en Red\n" ) ;
  printf( "       Su uso es:\n\tsumus [-w PuertoWeb] [-p PuertoMus] [-m MaxPartidas] [-d]
\t( -d indica que se sirvan paginas web del subdirectorio \"web\".)\n\n" ) ;
  exit( 0 ) ;
 }

 printf( "Programa servidor de partidas de mus. ( sumus v0.2.1 )\n") ;
 printf( "\t(C) 2000 Antonio Garcia (angarcia@openbank.es)\n\n" ) ;

 /* LLEGA UN MOMENTO EN LA HISTORIA DE CADA PROGRAMA EN EL QUE DEBE 
    RESERVAR MEMORIA PARA SUS ESTRUCTURAS DE DATOS E INICIALIZARLAS */
  Partidas = (struct sPartida *)calloc( MaxPartidas, sizeof( struct sPartida )) ;
  if( Partidas == NULL )
   {
    printf( "Error: Imposible asignar memoria para %d partidas.\n", MaxPartidas ) ;
    exit(-1) ;
   }

 /* ESTE ES UN BUEN MOMENTO PARA REGISTRAR LOS sockets Y ESCUCHAR LOS PUERTOS
    WEB y MUS */
  SocketWeb = socket( AF_INET, SOCK_STREAM, 0 ) ;    
  SocketMus = socket( AF_INET, SOCK_STREAM, 0 ) ;
  if( SocketWeb < 0 || SocketMus < 0 )
   {
     printf( "Error: Imposible asignar recursos de comunicaciones (sockets)\n" ) ;
     exit( -2 ) ;
   }

 /* AHORA LOS HAGO NO BLOQUEANTES */
  if (  ioctl( SocketWeb, FIONBIO, &bloquea ) < 0 ||
        ioctl( SocketMus, FIONBIO, &bloquea ) < 0 )
  {
   printf( "Error: Este sistema no me permite usar comunicaciones asincronas.\n" ) ;
   exit ( -3 ) ;
  }

 /* Y AHORA LOS PREPARO PARA ESCUCHAR PETICIONES WEB Y MUS */    
  memset( &host, 0, sizeof(host) ) ;
  host.sin_family = AF_INET ;
  host.sin_port = htons( PuertoWeb ) ;
  host.sin_addr.s_addr = htonl(INADDR_ANY) ;
  if( bind( SocketWeb, (struct sockaddr *) &host, sizeof( host ) ) )
   {
     printf( "Error: Sistema de comunicaciones (bind Web).\n" ) ;
     exit( -3 ) ;
   }
  if( listen( SocketWeb, 50 ) )
   {
     printf( "Error: Sistema de comunicaciones (listen Web).\n" ) ;
     exit( -4 ) ;
   }

  memset( &host, 0, sizeof(host) ) ;
  host.sin_family = AF_INET ;
  host.sin_port = htons( PuertoMus ) ;
  host.sin_addr.s_addr = htonl(INADDR_ANY) ;
  if( bind( SocketMus, (struct sockaddr *) &host, sizeof( host ) ) )
   {
     printf( "Error: Sistema de comunicaciones (bind Mus).\n" ) ;
     exit( -5 ) ;
   }
  if( listen( SocketMus, MaxPartidas * 4 + 20 ) )
   {
     printf( "Error: Sistema de comunicaciones (listen Mus).\n" ) ;
     exit( -6 ) ;
   }

 printf( "Acepta un maximo de %d partidas simultaneas\nPuerto conexiones de mus: %d\n",MaxPartidas,PuertoMus ) ;
 printf( "Puerto informacin Web: %d\n",PuertoWeb ) ;
 printf( "Preparado para recibir peticiones!\n" ) ;
 /* Y EN ESTE BUCLE ESTA LA MADRE DEL CORDERO DEL PROGRAMA
    SE TRATA DE ESCUCHAR LAS PETICIONES DE CONEXION Y LOS MENSAJES DE
    LOS QUE YA ESTAN CONECTADOS EN sockets NO BLOQUEANTES Y RESPONDER
    DE PRISA. ESO ME PERMITIRA LIBRARME DE TRABAJAR CON threads Y forks */
 while( 1 /*true*/ )
 {
   sleep( 1 ) ; /* PARA NO CONSUMIR TODA LA CPU DE LA MAQUINA */
   /* PRIMERO VER SI HAY CONEXIONES WEB NUEVAS */
  NuevoSocket = accept( SocketWeb, NULL, 0 ) ;
  if( NuevoSocket >= 0 )
    PonEnListaWeb( NuevoSocket ) ;
  else
    if( errno != EAGAIN && errno != EWOULDBLOCK )
      {
        printf( "Error: No puedo Seguir escuchando peticiones Web\n." ) ;
        exit( -7 ) ;
      }
   /* SEGUNDO VER SI HAY CONEXIONES MUS NUEVAS */
  NuevoSocket = accept( SocketMus, NULL, 0 ) ;
  if( NuevoSocket >= 0 )
    PonEnListaMus( NuevoSocket ) ;
  else
    if( errno != EAGAIN && errno != EWOULDBLOCK )
      {
        printf( "Error: No puedo Seguir escuchando peticiones Mus\n." ) ;
        exit( -8 ) ;
      }
   /* TERCERO VER SI HAY MENSAJES NUEVOS DE LAS CONEXIONES WEB */
  for( j = CuantosWebPendientes-1 ; j >= 0 ; j-- )
  {
   k = recv( SocketWebPendientes[ j ], Buffer, 20480, 0 ) ;
   if( k > 0 )
      RespondeHTTPPendiente( j ) ;
   else
    if( (errno != EAGAIN && errno != EWOULDBLOCK) || k == 0 )
     {
       close( SocketWebPendientes[ j ] ) ;
       SacaDeListaWeb( j ) ;
     }
  }
   /* CUARTO VER SI HAY MENSAJES NUEVOS DE LAS CONEXIONES MUS NUEVAS */
  for( j = CuantosMusPendientes-1 ; j >= 0 ; j-- )
   {
     k = recv( SocketMusPendientes[ j ], Buffer, 20480, 0 ) ;
     if(  k > 0 )
       RespondeMusPendiente( j ) ;
     else
      if( (errno != EAGAIN && errno != EWOULDBLOCK) || k == 0 )
      {
       close( SocketMusPendientes[ j ] ) ;
       SacaDeListaMus( j ) ;
      }
   }
   /* QUINTO Y ULTIMO VER SI HAY MENSAJES NUEVOS DE LAS PARTIDAS DE MUS */
  Ahora = time( NULL ) ;
  j = PosPrimeraPartida ;
  while( j > -1 )
  {
   if( Ahora - Partidas[j].UltimaAccion > TIMEOUT )
    {
     printf( "Partida %d [%s] TimeOut. Se cierra.\n", j, Partidas[j].NombrePartida ) ;
     for( k=0 ; k<4 ; k++)
       if( Partidas[j].Canutos[k] > -1 )
         Envia( Partidas[j].Canutos[k], "BTimeOut. No hubo comunicacion en 10 minutos." ) ;
     k = Partidas[j].Siguiente ;
     EliminaPartida( j ) ;
     j = k ;
    }
   else
    {
     l = j ;
     j = Partidas[j].Siguiente ; /* TrataMensaje puede hacer que desaparezca el registro */
     for( k=0 ; k<4 ; k++)
       if( Partidas[l].Canutos[k] > -1 )
         {
          p = recv( Partidas[l].Canutos[k], Buffer, sizeof( struct sMensaje ),0 ) ;
          if( p > 0 )
             TrataMensaje( l, k, Buffer ) ;
          else
           if( p == 0 ) /* Este cobarde cerro la conexion sin despedirse. */
              TrataMensaje( l, k, "F" ) ;
         }
    }
  }
 }
/* EN REALIDAD DE ESTE BUCLE NO SE SALE NUNCA */  

/* NO SE PORQUE HAGO ESTO, YA QUE EL PROGRAMA SOLO SE TERMINA CON UN CTL+C
   PERO QUEDA BIEN DE TODAS FORMAS */   
free( Partidas ) ;
return 1 ;
}

/******* FIN DEL FICHERO SUMUS.C *********/
